/* dhcpc.c
 *
 * udhcp DHCP client
 *
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/sysinfo.h>
#include <unistd.h>
#include <getopt.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <time.h>
#include <string.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <errno.h>

#include "dhcpd.h"
#include "dhcpc.h"
#include "options.h"
#include "clientpacket.h"
#include "packet.h"
#include "script.h"
#include "socket.h"
#include "debug.h"
#include "pidfile.h"
#include "nlhelp.h"
#include "zeroconf.h"

#include "csman.h"
#include "unilog.h"
#include "csid/csid_dhcp.h"
static int state;
static unsigned long requested_ip; /* = 0 */
static unsigned long server_addr;
static time_t timeout;
static int packet_num; /* = 0 */
static int fd = -1;
static int signal_pipe[2];
static struct rtnl_handle nlh;

#define LISTEN_NONE 0
#define LISTEN_KERNEL 1
#define LISTEN_RAW 2
#define LISTEN_DEMAND 3
static int listen_mode;

#define DEFAULT_SCRIPT	"/usr/share/udhcpc/default.script"
#define DEMAND_FILE		"/var/run/udhcpc.demand"

static struct zc_config zc;

/* this is the time at which 'aggressive' search for a DHCP server ends */
static time_t time_aggressive_end;

/* this is the duration, in seconds, of the 'aggressive' search */
#define AGGR_DURATION  120

/* this is the retry frequency of the 'aggressive' search, in seconds */
#define AGGR_INTERVAL    5

/* this is the retry frequency of the 'normal' search, in seconds */
#define NORMAL_INTERVAL 60

struct client_config_t client_config = {
    /* Default options. */
    abort_if_no_lease: 0,
    foreground: 0,
    quit_after_lease: 0,
    background_if_no_lease: 0,
    interface: "eth0",
    pidfile: NULL,
    wifi: NULL,
    script: DEFAULT_SCRIPT,
    clientid: NULL,
    hostname: NULL,
    ifindex: 0,
    demand: 0,
    lanif: "br0",
    arp: "\0\0\0\0\0\0",		/* appease gcc-3.0 */
};

#ifndef BB_VER
static void show_usage(void)
{
    printf(
           "Usage: udhcpc [OPTIONS]\n\n"
           "  -c, --clientid=CLIENTID         Client identifier\n"
           "  -H, --hostname=HOSTNAME         Client hostname\n"
           "  -h                              Alias for -H\n"
           "  -f, --foreground                Do not fork after getting lease\n"
           "  -b, --background                Fork to background if lease cannot be\n"
           "                                  immediately negotiated.\n"
           "  -i, --interface=INTERFACE       Interface to use (default: eth0)\n"
           "  -n, --now                       Exit with failure if lease cannot be\n"
           "                                  immediately negotiated.\n"
           "  -p, --pidfile=file              Store process ID of daemon in file\n"
           "  -q, --quit                      Quit after obtaining lease\n"
           "  -r, --request=IP                IP address to request (default: none)\n"
           "  -s, --script=file               Run file at dhcp events (default:\n"
           "                                  " DEFAULT_SCRIPT ")\n"
           "  -w, --wifi                      Interface for SSID change monitor\n"
           "  -v, --version                   Display version\n"
           "  -d, --demand					  Connect on Demand\n"					  	
           "  -l, --lanif					  lan interface name\n"						  	
           );
    exit(0);
}
#endif


/* just a little helper */
static void change_mode(int new_mode)
{
    DEBUG(LOG_INFO, "entering %s listen mode",
          new_mode ? (new_mode == 1 ? "kernel" : "raw") : "none");
    if (fd != -1) {
        close(fd);
        fd = -1;
    }
    listen_mode = new_mode;
}


/* perform a renew */
static void perform_renew(void)
{
    int state_before = state;

    LOG(LOG_INFO, "Performing a DHCP renew");
    switch (state) {
    case BOUND:
        change_mode(LISTEN_KERNEL);
    case RENEWING:
    case REBINDING:
        state = RENEW_REQUESTED;
        break;
    case RENEW_REQUESTED: /* impatient are we? fine, square 1 */
        run_script(NULL, "deconfig");
    case REQUESTING:
    case RELEASED:
        change_mode(LISTEN_RAW);
        state = INIT_SELECTING;
        break;
    case INIT_SELECTING:
        break;
    }

    /* start things over */
    packet_num = 0;

    /* Kill any timeouts because the user wants this to hurry along */
    timeout = 0;

    if (state_before == INIT_SELECTING) {
        /* If we haven't ever seen a DHCP server, and we are happily using a
           zeroconf adress, then 'renew' the zeroconf-assigned address, but 
           don't release it.  This avoids unnecessary IP address churn. */
        switch(zc.zc_state) {
        case ZEROCONF_DISABLED:
        case ZEROCONF_PROBING:
        default:
            zc_reset(&zc);
            break;
        case ZEROCONF_ANNOUNCING:
        case ZEROCONF_RUNNING:
            zc.zc_state = ZEROCONF_ANNOUNCING;
            zc.zc_count = 0;
            zc.zc_timeout = 0;
            break;
        }
    } else
        zc_reset(&zc);
}


/* perform a release */
static void perform_release(void)
{
    char buffer[16];
    struct in_addr temp_addr;

    /* send release packet */
    if (state == BOUND || state == RENEWING || state == REBINDING) {
        temp_addr.s_addr = server_addr;
        sprintf(buffer, "%s", inet_ntoa(temp_addr));
        temp_addr.s_addr = requested_ip;
        LOG(LOG_INFO, "Unicasting a release of %s to %s", 
            inet_ntoa(temp_addr), buffer);
        send_release(server_addr, requested_ip); /* unicast */
        run_script(NULL, "deconfig");
    }
    LOG(LOG_INFO, "Entering released state");

    change_mode(LISTEN_NONE);
    state = RELEASED;
    timeout = 0x7fffffff;

    /* make sure zeroconf isn't running */
    zc_reset(&zc);
}


/* Exit and cleanup */
static void exit_client(int retval)
{
    pidfile_delete(client_config.pidfile);
    CLOSE_LOG();
    exit(retval);
}


/* Signal handler */
static void signal_handler(int sig)
{
    if (send(signal_pipe[1], &sig, sizeof(sig), MSG_DONTWAIT) < 0) {
        LOG(LOG_ERR, "Could not send signal: %s",
            strerror(errno));
    }
}


void background(void)
{
    int pid_fd;

    /* hold lock during fork. */
    pid_fd = pidfile_acquire(client_config.pidfile); 
    /* don't let daemon close it */
    while (pid_fd >= 0 && pid_fd < 3) pid_fd = dup(pid_fd); 
    if (daemon(0, 0) == -1) {
        perror("fork");
        exit_client(1);
    }
    client_config.foreground = 1; /* Do not fork again. */
    pidfile_write_release(pid_fd);
}

static int max_of_four(int w, int x, int y, int z)
{
    int max = w;

    if (x > max)
        max = x;
    if (y > max)
        max = y;
    if (z > max)
        max = z;

    return max;
}

#ifdef COMBINED_BINARY
int udhcpc_main(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{
    unsigned char *temp, *message;
    unsigned long t1 = 0, t2 = 0, xid = 0;
    unsigned long start = 0, lease,lease_csid;
    fd_set rfds;
    int retval;
    struct timeval tv;
    int c, len;
    struct dhcpMessage packet;
    struct in_addr temp_addr;
    int pid_fd;
    time_t now;
    int max_fd;
    int sig;
    u_int32_t lan_ip;
    u_int32_t lan_mask;
    char buf[200];
    struct sysinfo info;
    int csman_fd;
	
    //int count;

    static struct option arg_options[] = {
        {"clientid",	required_argument,	0, 'c'},
        {"foreground",	no_argument,		0, 'f'},
        {"background",	no_argument,		0, 'b'},
        {"hostname",	required_argument,	0, 'H'},
        {"hostname",    required_argument,      0, 'h'},
        {"interface",	required_argument,	0, 'i'},
        {"now", 	no_argument,		0, 'n'},
        {"pidfile",	required_argument,	0, 'p'},
        {"quit",	no_argument,		0, 'q'},
        {"request",	required_argument,	0, 'r'},
        {"script",	required_argument,	0, 's'},
        {"wifi",        required_argument,      0, 'w'},
        {"version",	no_argument,		0, 'v'},
        {"help",	no_argument,		0, '?'},
        {"demand", no_argument,			0, 'd'},
        {"lanif", required_argument, 0, 'l'},
        {0, 0, 0, 0}
    };

    /* get options */
    while (1) {
        int option_index = 0;
        c = getopt_long(argc, argv, "c:fbH:h:i:np:qr:s:w:vdl:", arg_options, 
                        &option_index);
        if (c == -1) break;
		
        switch (c) {
        case 'c':
            len = strlen(optarg) > 255 ? 255 : strlen(optarg);
            if (client_config.clientid) free(client_config.clientid);
            client_config.clientid = xmalloc(len + 2);
            client_config.clientid[OPT_CODE] = DHCP_CLIENT_ID;
            client_config.clientid[OPT_LEN] = len;
            client_config.clientid[OPT_DATA] = '\0';
            strncpy(client_config.clientid + OPT_DATA, optarg, len);
            break;
        case 'f':
            client_config.foreground = 1;
            break;
        case 'b':
            client_config.background_if_no_lease = 1;
            break;
        case 'h':
        case 'H':
            len = strlen(optarg) > 255 ? 255 : strlen(optarg);
            if (client_config.hostname) free(client_config.hostname);
            client_config.hostname = xmalloc(len + 2);
            client_config.hostname[OPT_CODE] = DHCP_HOST_NAME;
            client_config.hostname[OPT_LEN] = len;
            strncpy(client_config.hostname + 2, optarg, len);
            break;
        case 'i':
            client_config.interface =  optarg;
            break;
        case 'n':
            client_config.abort_if_no_lease = 1;
            break;
        case 'p':
            client_config.pidfile = optarg;
            break;
        case 'q':
            client_config.quit_after_lease = 1;
            break;
        case 'r':
            requested_ip = inet_addr(optarg);
            break;
        case 's':
            client_config.script = optarg;
            break;
        case 'w':
            client_config.wifi = optarg;
            break;
        case 'v':
            printf("udhcpcd, version %s\n\n", VERSION);
            exit_client(0);
            break;
        case 'd':
        	client_config.demand=1;
        	break;    
        case 'l':        	
        	client_config.lanif=optarg;        	
        	break;    
        default:
            show_usage();
        }
    }

    OPEN_LOG("udhcpc");
    LOG(LOG_INFO, "udhcp client (v%s) started", VERSION);

    pid_fd = pidfile_acquire(client_config.pidfile);
    pidfile_write_release(pid_fd);

    if (read_interface(client_config.interface, &client_config.ifindex, 
                       NULL, client_config.arp) < 0)
        exit_client(1);
		
    if (!client_config.clientid) {
        client_config.clientid = xmalloc(6 + 3);
        client_config.clientid[OPT_CODE] = DHCP_CLIENT_ID;
        client_config.clientid[OPT_LEN] = 7;
        client_config.clientid[OPT_DATA] = 1;
        memcpy(client_config.clientid + 3, client_config.arp, 6);
    }

    /* if appropriate, set up to listen for SSID change notifications. */
    if (client_config.wifi) {
#ifdef __SH4__
        int ret = rtnl_open(&nlh, RTMGRP_Rincon);
#else
        int ret = -1;
#endif
        if (ret != 0) {
            LOG(LOG_ERR, "FATAL: couldn't listen for SSID change, %s", 
                strerror(errno));
            exit_client(0);                
        }
    } else
        nlh.fd = -1;

    /* setup signal handlers */
    socketpair(AF_UNIX, SOCK_STREAM, 0, signal_pipe);
    signal(SIGUSR1, signal_handler);
    signal(SIGUSR2, signal_handler);
    signal(SIGTERM, signal_handler);
	
    state = INIT_SELECTING;
    zc_init(&zc, client_config.arp);
    run_script(NULL, "deconfig");
	
	if( client_config.demand){
		char rgch[32];
			
        zc_start_probing(&zc, 0x7fffffff);	
        sprintf(rgch, "%u.%u.%u.%u", zc.zc_addr[0], zc.zc_addr[1], zc.zc_addr[2], zc.zc_addr[3]);
        run_script_zeroconf(rgch, "255.255.0.0", "bound");      
		change_mode(LISTEN_DEMAND);
		sprintf(buf, "echo 'demand' > %s", DEMAND_FILE);
		system(buf);
	}else{	
    change_mode(LISTEN_RAW);
	}

    time_aggressive_end = time(0) + AGGR_DURATION;

	//count=0;
    for (;;) {
    	    	
    	//count++;    	
    	//syslog(LOG_INFO," count=%d  timeout=%d zc_timeout=%d , time(0)=%d, tv_sec=%d \n", count, 
    	//timeout,zc.zc_timeout, time(0), ((zc.zc_timeout>timeout)?(timeout - time(0)):(zc.zc_timeout - time(0)) ));
    	
        /* we are juggling two state machines in parallel here; set the
           timeout for the select to the sooner of the DHCP and 
           zeroconf timeouts. */
        if (zc.zc_timeout > timeout)
            tv.tv_sec = timeout - time(0);
        else
            tv.tv_sec = zc.zc_timeout - time(0);
                
        tv.tv_usec = 0;
        FD_ZERO(&rfds);

        if (listen_mode != LISTEN_NONE && fd < 0) {
            if (listen_mode == LISTEN_KERNEL)
                fd = listen_socket(INADDR_ANY, CLIENT_PORT, 
                                   client_config.interface);
            else if ( listen_mode == LISTEN_RAW )
                fd = raw_socket(client_config.ifindex);
            else if( listen_mode == LISTEN_DEMAND)
            	fd=raw_demand_socket(client_config.lanif, &lan_ip, &lan_mask);
            	   
            if (fd < 0) {
                LOG(LOG_ERR, "FATAL: couldn't listen on socket, %s", 
                    strerror(errno));
                exit_client(0);
            }
        }

        if (zc.zc_state != ZEROCONF_DISABLED && zc.zc_fd < 0) {
            zc.zc_fd = zc_arp_socket(client_config.interface);
            if (zc.zc_fd < 0) {
                LOG(LOG_ERR, "FATAL: couldn't listen on ARP socket, %s", 
                    strerror(errno));
                exit_client(0);                        
            }
        }

        if (fd >= 0) FD_SET(fd, &rfds);
        if (zc.zc_fd >= 0) FD_SET(zc.zc_fd, &rfds);
        if (nlh.fd >= 0) FD_SET(nlh.fd, &rfds);
        FD_SET(signal_pipe[0], &rfds);

        if (tv.tv_sec >=0) {
            max_fd = max_of_four(fd, zc.zc_fd, nlh.fd, signal_pipe[0]);
            if( listen_mode == LISTEN_DEMAND ){
            	retval = select(max_fd + 1, &rfds, NULL, NULL, NULL);
        	}else{
            retval = select(max_fd + 1, &rfds, NULL, NULL, &tv);
            }	
        } else 
            retval = 0; /* If we already timed out, fall through */

        now = time(0);
                
        /* time-driven processing for DHCP state machine. */
        if (retval == 0 && timeout <= now) {
            switch (state) {
            case INIT_SELECTING:
            
            	if( listen_mode == LISTEN_DEMAND){
            		timeout=time(0);
            		background();
            		goto out1;
            	}	
            
                /* make sure we're listenting to all IP datagrams; if not,
                   start and get us to be called again (i.e. don't change 
                   the value of timeout. */
                if (listen_mode != LISTEN_RAW) {
                    change_mode(LISTEN_RAW);
                    goto out1;
                }

                if (packet_num < 3) {
                    if (packet_num == 0)
                        xid = random_xid();

                    /* send discover packet */
                    send_discover(xid, requested_ip); /* broadcast */
					
                    timeout = now + ((packet_num == 2) ? 4 : 2);
                    packet_num++;
                } else {
                    if (client_config.background_if_no_lease) {
                        LOG(LOG_INFO, "No lease, forking to background.");
                        background();
                    } else if (client_config.abort_if_no_lease) {
                        LOG(LOG_INFO, "No lease, failing.");
                        exit_client(1);
                    }
                    /* wait to try again */
                    packet_num = 0;
                    if (now < time_aggressive_end)
                        timeout = now + AGGR_INTERVAL;
                    else
                        timeout = now + NORMAL_INTERVAL;

                    /* while waiting for the next DHCP attempt, don't bother
                       listening on the DHCP socket at all; it's just a waste
                       of resources since the socket gets all IP datagrams. */
                    change_mode(LISTEN_NONE);

                    /* in the meantime, run zeroconf, if it isn't already */
                    if (zc.zc_state == ZEROCONF_DISABLED)
                        zc_start_probing(&zc, now);
                }
                break;
            case RENEW_REQUESTED:
            case REQUESTING:
                if (packet_num < 3) {
                    /* send request packet */
                    if (state == RENEW_REQUESTED)
                        send_renew(xid, server_addr, requested_ip); /* ucast */
                    else 
                        send_selecting(xid, server_addr, requested_ip); /*bc*/
					
                    timeout = now + ((packet_num == 2) ? 10 : 2);
                    packet_num++;
                } else {
                    /* timed out, go back to init state */
                    if (state == RENEW_REQUESTED) run_script(NULL, "deconfig");
                    state = INIT_SELECTING;
                    timeout = now;
                    packet_num = 0;
                    change_mode(LISTEN_RAW);
                }
                break;
            case BOUND:
                /* Lease is starting to run out, enter renewing state */
                state = RENEWING;
                change_mode(LISTEN_KERNEL);
                DEBUG(LOG_INFO, "Entering renew state");
                /* fall right through */
            case RENEWING:
                /* Either set a new T1, or enter REBINDING state */
                if ((t2 - t1) <= (lease / 14400 + 1)) {
                    /* timed out, enter rebinding state */
                    state = REBINDING;
                    timeout = now + (t2 - t1);
                    DEBUG(LOG_INFO, "Entering rebinding state");
                } else {
                    /* send a request packet */
                    send_renew(xid, server_addr, requested_ip); /* unicast */
					
                    t1 = (t2 - t1) / 2 + t1;
                    timeout = t1 + start;
                }
                break;
            case REBINDING:
                /* Either set a new T2, or enter INIT state */
                if ((lease - t2) <= (lease / 14400 + 1)) {
                    /* timed out, enter init state */
                    state = INIT_SELECTING;
                    LOG(LOG_INFO, "Lease lost, entering init state");
                    run_script(NULL, "deconfig");
                    timeout = now;
                    packet_num = 0;
                    change_mode(LISTEN_RAW);
                } else {
                    /* send a request packet */
                    send_renew(xid, 0, requested_ip); /* broadcast */

                    t2 = (lease - t2) / 2 + t2;
                    timeout = t2 + start;
                }
                break;
            case RELEASED:
                /* yah, I know, *you* say it would never happen */
                timeout = 0x7fffffff;
                break;
            }
        out1:
            ;
        } 

        /* I/O-driven processing for DHCP state machine. */
        if (retval > 0 && listen_mode != LISTEN_NONE && FD_ISSET(fd, &rfds)) {
            if (listen_mode == LISTEN_KERNEL){
                len = get_packet(&packet, fd);
            }else if( listen_mode == LISTEN_RAW){
                len = get_raw_packet(&packet, fd);
            }else if( listen_mode == LISTEN_DEMAND){
            	retval=check_raw_demand_packet(fd, lan_ip, lan_mask);
	        	if(retval<0){
	        		LOG(LOG_ERR,"FATAL: check_raw_packet_for_demand failed");
	        		exit_client(1); 
	        	}else if( retval>0){ // OK!!	        		
        			zc_reset(&zc);
        			timeout=0;
        			change_mode(LISTEN_RAW);
        			time_aggressive_end = time(0) + AGGR_DURATION;
        			sprintf(buf, "echo 'raw' > %s", DEMAND_FILE);
					system(buf);
	        		goto out2;
	        	}else if( retval==0){ // Failed!!	        		
	        		timeout=time(0);	
	        		goto out2;
	        	}				
            }	     
			
            if (len == -1 && errno != EINTR) {
                DEBUG(LOG_INFO, "error on read, %s, reopening socket", 
                      strerror(errno));
                change_mode(listen_mode); /* just close and reopen */
            }
            if (len < 0) 
                goto out2;
			
            if (packet.xid != xid) {
                DEBUG(LOG_INFO, "Ignoring XID %lx (our xid is %lx)",
                      (unsigned long) packet.xid, xid);
                goto out2;
            }
			
            if ((message = get_option(&packet, DHCP_MESSAGE_TYPE)) == NULL) {
                DEBUG(LOG_ERR, "couldnt get option from packet -- ignoring");
                goto out2;
            }
			
            switch (state) {
            case INIT_SELECTING:
                /* Must be a DHCPOFFER to one of our xid's */
                if (*message == DHCPOFFER) {
                    if ((temp = get_option(&packet, DHCP_SERVER_ID))) {
                        memcpy(&server_addr, temp, 4);
                        xid = packet.xid;
                        requested_ip = packet.yiaddr;
						
                        /* enter requesting state */
                        state = REQUESTING;
                        timeout = now;
                        packet_num = 0;
                    } else {
                        DEBUG(LOG_ERR, "No server ID in message");
                    }
                }
                break;
            case RENEW_REQUESTED:
            case REQUESTING:
            case RENEWING:
            case REBINDING:
            
                if (*message == DHCPACK) {               	
                    if (!(temp = get_option(&packet, DHCP_LEASE_TIME))) {
                        LOG(LOG_ERR, 
                            "No lease time with ACK, using 1 hour lease");
                        lease = 60 * 60;
                    } else {
                        memcpy(&lease, temp, 4);
                        lease = ntohl(lease);                                                
                    }
                    
                	sysinfo(&info);
                    csman_fd=open_csman(NULL,0);
                    lease_csid=info.uptime+lease;
                    write_csman(csman_fd, CSID_S_DHCP_CLI_LEASE_TIME, (const void *)&lease_csid, sizeof(unsigned long), 0);
					close_csman(csman_fd);	
					LPRINTF(PRI_D1, " lease_csid=%ld uptime=%ld lease=%ld\n", lease_csid, info.uptime, lease);
                    /* make sure zeroconf isn't running */
                    zc_reset(&zc);
						
                    /* enter bound state */
                    t1 = lease / 2;
					
                    /* little fixed point for n * .875 */
                    t2 = (lease * 0x7) >> 3;
                    temp_addr.s_addr = packet.yiaddr;
                    LOG(LOG_INFO, "Lease of %s obtained, lease time %ld", 
                        inet_ntoa(temp_addr), lease);
                    start = now;
                    timeout = t1 + start;
                    requested_ip = packet.yiaddr;
                    run_script(&packet,
                               ((state == RENEWING || state == REBINDING) ? 
                                "renew" : "bound"));

                    state = BOUND;
                    change_mode(LISTEN_NONE);
                    if (client_config.quit_after_lease) 
                        exit_client(0);
                    if (!client_config.foreground)
                        background();
                } else if (*message == DHCPNAK) {
                    /* return to init state */
                    LOG(LOG_INFO, "Received DHCP NAK");
                    run_script(&packet, "nak");
                    if (state != REQUESTING)
                        run_script(NULL, "deconfig");
                    state = INIT_SELECTING;
                    timeout = now;
                    requested_ip = 0;
                    packet_num = 0;
                    change_mode(LISTEN_RAW);
                    sleep(3); /* avoid excessive network traffic */
                }
                break;
                /* case BOUND, RELEASED: - ignore all packets */
            }
        out2:
            ;
        }

        /* time-driven processing for zeroconf state machine. */
        if (retval == 0 && zc.zc_timeout <= now)
            zc_handle_timer(&zc, now);

        /* I/O-driven processing for zeroconf state machine. */
        if (retval > 0 && zc.zc_fd != -1 && FD_ISSET(zc.zc_fd, &rfds))
            zc_handle_input(&zc, now);

        /* processing of netlink messages */
        if (retval > 0 && nlh.fd != -1 && FD_ISSET(nlh.fd, &rfds)) {
            if (rtnl_msg_triggers_dhcp_renew(&nlh, client_config.wifi)) {
                /* nlhelp.c determined that this message should trigger a
                   DHCP renew; do it. */
                time_aggressive_end = time(0) + AGGR_DURATION;
                perform_renew();
            }
        }

        /* signal-driven processing */
        if (retval > 0 && FD_ISSET(signal_pipe[0], &rfds)) {
            if (read(signal_pipe[0], &sig, sizeof(sig)) < 0) {
                DEBUG(LOG_ERR, "Could not read signal: %s", 
                      strerror(errno));
                goto out3; /* probably just EINTR */
            }
            switch (sig) {
            case SIGUSR1: 
                perform_renew();
                break;
            case SIGUSR2:
                perform_release();
                break;
            case SIGTERM:
                LOG(LOG_INFO, "Received SIGTERM");
                exit_client(0);
            }
        out3:
            ;
        }
    }
    return 0;
}

